/*
	(C) 1999-2000 AROS - The Amiga Research OS
	$Id: conclip.c,v 1.2 2005/04/11 20:00:33 piru Exp $
 
	Desc: 
	Lang: English
*/

/*****************************************************************************************/

#include <exec/memory.h>
#include <exec/execbase.h>
#include <dos/dos.h>
#include <dos/exall.h>
#include <dos/dostags.h>
#include <intuition/intuition.h>
#include <intuition/sghooks.h>
#include <intuition/classusr.h>
#include <libraries/iffparse.h>
#include <devices/inputevent.h>
#include <devices/console.h>
#include <devices/conunit.h>
#include <datatypes/textclass.h>
#include <utility/hooks.h>

#include <aros/asmcall.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/iffparse.h>
#include <proto/alib.h>
#include <proto/utility.h>
#include <proto/console.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#undef DEBUG
#define DEBUG 0
#include <aros/debug.h>

static void ConClipHandler(void);
static void	SendBreak(void);

#include "conclip_version.h"

static const char version[] = VERSTAG;

/*****************************************************************************************/

#define ARG_TEMPLATE 	"CLIPUNIT=UNIT/N,OFF/S"
#define ARG_CLIPUNIT 	0
#define ARG_OFF		1
#define NUM_ARGS	2

/*****************************************************************************************/

#define EDITCODE_COPY	'C'
#define EDITCODE_PASTE	'V'
#define SNIPCODE_COPY	0x100

struct MyEditHookMsg
{
	struct Message 	msg;
	struct SGWork	*sgw;
	WORD 		code;
	char		*snipdata;
	ULONG		sniplength;
};

/*****************************************************************************************/

const STRPTR CONCLIP_TASKNAME = " ConClip ";
const STRPTR CONCLIP_PORTNAME = "ConClip.rendezvous";

/*****************************************************************************************/

static struct MsgPort 	*progport;
static struct Hook	edithook, *oldedithook, sniphook;
static ULONG 		clipunit, portmask;
static BOOL 		off;
static UBYTE 		s[256];
struct ExecBase		*SysBase;
struct DosLibrary	*DOSBase;
struct Library		*UtilityBase;
struct IntuitionBase	*IntuitionBase;
struct Library		*IFFParseBase;
struct Library		*ConsoleDevice;
BPTR			OldSegList;
struct IOStdReq 	snipio;

/*****************************************************************************************/

static void installedithook(void);

int main(void)
{
	struct DosLibrary	*DOSBase;
	struct Process		*MyProcess;
	struct RDArgs		*myargs;
	IPTR 			args[NUM_ARGS] = {0, 0};

	SysBase=*((struct ExecBase**) 4);
	if (DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37))
	{
		if (myargs = ReadArgs(ARG_TEMPLATE, args, NULL))
		{
			if (args[ARG_CLIPUNIT])
			{
				clipunit = (ULONG)(*(IPTR *)args[ARG_CLIPUNIT]);
			}
			if (args[ARG_OFF])
			{
				off = args[ARG_OFF] ? TRUE : FALSE;
			}

			if (off)
			{
				SendBreak();
			}
			else
			{
				struct CommandLineInterface *cli;

				cli			=	Cli();
				OldSegList		=	cli->cli_Module;

				if (MyProcess=CreateNewProcTags(NP_Entry,
				                                (ULONG)  &ConClipHandler,
				                                NP_Name, (ULONG) CONCLIP_TASKNAME,
				                                NP_Priority, 0,
				                                NP_CodeType, MACHINE_PPC,
				                                NP_ConsoleTask, NULL,
				                                TAG_END))
				{
					cli->cli_Module	=	NULL;

					D(bug("ConClip: Process 0x%lx\n",
						  (ULONG)MyProcess));
				}
			}
			FreeArgs(myargs);
		}
		else
		{
			Fault(IoErr(), 0, s, 255);
		}
		CloseLibrary((struct Library*) DOSBase);
	}
	return 0;
}


/*****************************************************************************************/

static void	SendBreak(void)
{
	struct MsgPort	*MyPort;
	Forbid();
	if (MyPort=FindPort(CONCLIP_PORTNAME))
	{
		Signal(MyPort->mp_SigTask, SIGBREAKF_CTRL_C);
	}
	Forbid();
}

static ULONG createport(void)
{
	Forbid();
	progport = FindPort(CONCLIP_PORTNAME);
	if (progport)
	{
		Permit();
		progport = NULL;
		return(FALSE);
	}
	progport = CreatePort(CONCLIP_PORTNAME, 1);
	Permit();

	if (!progport)
	{
		D(bug("ConClip: Could not create MsgPort!\n"));
		return(FALSE);
	}
	portmask = 1L << progport->mp_SigBit;
	return(TRUE);
}

/*****************************************************************************************/
/*****************************************************************************************/
/*****************************************************************************************/

static void savetoclipboard(char *Data,
							ULONG Length)
{
	struct IFFHandle *iff;

	D(bug("ConClip/savetoclipboard: Data 0x%lx Length %ld\n",
		  Data,
		  Length));
	if((iff = AllocIFF()))
	{
		D(bug("ConClip/savetoclipboard: iff 0x%lx\n",
			  iff));
		if((iff->iff_Stream = (IPTR)OpenClipboard(clipunit)))
		{
			D(bug("ConClip/savetoclipboard: iffstream 0x%lx\n",
				  iff->iff_Stream));
			InitIFFasClip(iff);
			if(!OpenIFF(iff,IFFF_WRITE))
			{
				D(bug("ConClip/savetoclipboard: IFFF_WRITE set\n"));
				if(!PushChunk(iff, ID_FTXT, ID_FORM, IFFSIZE_UNKNOWN))
				{
					D(bug("ConClip/savetoclipboard: pushed FORM\n"));
					if(!PushChunk(iff, ID_FTXT, ID_CHRS, IFFSIZE_UNKNOWN))
					{
						D(bug("ConClip/savetoclipboard: pushed CHRS\n"));
						WriteChunkBytes(iff, Data, Length);
						D(bug("ConClip/savetoclipboard: pushed Text size %ld\n",
							  Length));
						PopChunk(iff);
						D(bug("ConClip/savetoclipboard: popped chunk\n"));
					} /* if(!PushChunk(iff, ID_FTXT, ID_CHRS, IFFSIZE_UNKNOWN)) */
					PopChunk(iff);
					D(bug("ConClip/savetoclipboard: popped chunk\n"));
				} /* if(!PushChunk(iff, ID_FTXT, ID_FORM, IFFSIZE_UNKNOWN)) */
				CloseIFF(iff);
				D(bug("ConClip/savetoclipboard: closed iff\n"));
			} /* if(!OpenIFF(iff,IFFF_WRITE)) */
			CloseClipboard((struct ClipboardHandle*)iff->iff_Stream);
			D(bug("ConClip/savetoclipboard: close clipboard\n"));
		} /* if((iff->iff_Stream = (IPTR)OpenClipboard(clipunit))) */
		FreeIFF(iff);
		D(bug("ConClip/savetoclipboard: free iff\n"));
	} /* if((iff = AllocIFF()))) */
	D(bug("ConClip/savetoclipboard: done\n"));
}

/*****************************************************************************************/

AROS_UFH3(ULONG, conclipeditfunc,
		  AROS_UFHA(struct Hook *,		hook,		A0),
		  AROS_UFHA(struct SGWork *,		sgw,		A2),
		  AROS_UFHA(ULONG *, 			command,	A1)
		 )
{
	AROS_USERFUNC_INIT

	struct MsgPort 	 *port, replyport;
	struct MyEditHookMsg msg;
	BOOL		 calloldhook = TRUE;
	ULONG 		 retcode = 0;

	switch (*command)
	{
	case SGH_KEY:
		D(bug("ConClip/conclipeditfunc: is SGH_KEY\n"));

		if (sgw->IEvent->ie_Qualifier & IEQUALIFIER_RCOMMAND)
		{
			D(bug("ConClip/conclipeditfunc: qualifier RCOMMAND okay\n"));

			D(bug("ConClip/conclipeditfunc: key = %c\n", toupper(sgw->Code)));
			switch(toupper(sgw->Code))
			{
			case 'C':
				if (!sgw->NumChars) break;
				/* fall through */

			case 'V':

				if ((port = FindPort(CONCLIP_PORTNAME)))
				{
					calloldhook = FALSE;

					replyport.mp_Node.ln_Type	= NT_MSGPORT;
					replyport.mp_Node.ln_Name 	= NULL;
					replyport.mp_Node.ln_Pri 	= 0;
					replyport.mp_Flags 		= PA_SIGNAL;
					replyport.mp_SigBit 	= SIGB_SINGLE;
					replyport.mp_SigTask 	= FindTask(NULL);
					NewList(&replyport.mp_MsgList);

					msg.msg.mn_Node.ln_Type 	= NT_MESSAGE;
					msg.msg.mn_ReplyPort 	= &replyport;
					msg.msg.mn_Length 		= sizeof(msg);

					msg.code = toupper(sgw->Code);
					msg.sgw  = sgw;

					if ((msg.code == EDITCODE_COPY) || (sgw->NumChars < sgw->StringInfo->MaxChars - 1))
					{
						D(bug("ConClip/conclipeditfunc: Msg Port 0x%lx\n", port));
						SetSignal(0, SIGF_SINGLE);
						PutMsg(port, &msg.msg);
						WaitPort(&replyport);
					}

					if (msg.code == EDITCODE_PASTE)
					{
						WORD len = strlen(sgw->WorkBuffer);

						D(bug("ConClip/conclipeditfunc: Paste len %ld\n", len));

						if (len != sgw->NumChars)
						{
							sgw->NumChars = len;
							sgw->EditOp   = EO_BIGCHANGE;
							sgw->Actions  = SGA_USE | SGA_REDISPLAY;

							retcode = 1;
						}

					} /* if (msg.code == EDITCODE_COPY) */

				} /* if ((port = FindPort(CONCLIP_PORTNAME))) */

				break;
			default:
				D(bug("ConClip/conclipeditfunc: unknown cmd <%lc>\n",
					  sgw->Code));
				break;

			} /* switch(toupper(sgw->Code)) */

		} /* if (sgw->IEvent->ie_Qualifier & IEQUALIFIER_RCOMMAND) */
		break;
	default:
		D(bug("ConClip/conclipeditfunc: unknown SGH_ 0x%lx>\n",
			  *command));
		break;

	} /* switch (*command) */

	if (calloldhook)
	{
		retcode = CallHookPkt(oldedithook, sgw, command);
	}
	D(bug("ConClip/conclipeditfunc: retcode 0x%lx\n", retcode));

	return retcode;

	AROS_USERFUNC_EXIT
}


/*****************************************************************************************/

static void installedithook(void)
{
	edithook.h_Entry = (HOOKFUNC)AROS_ASMSYMNAME(conclipeditfunc);
	edithook.h_SubEntry = NULL;
	edithook.h_Data = NULL;

	oldedithook = SetEditHook(&edithook);
	D(bug("installedithook: oldedithook 0x%lx\n",
		  oldedithook));
}

/*****************************************************************************************/

/* Struct passed to hooks in A1 register - Piru
*/
struct conhookmessage
{
	ULONG			chm_null0;	/* always passed as 0 */
	ULONG			chm_sniplen;
	UBYTE			*chm_snipdata;
};

AROS_UFH3(ULONG, conclipsnipfunc,
		  AROS_UFHA(struct Hook *,		hook,		A0),
		  AROS_UFHA(void *,			dummy,		A2),
		  AROS_UFHA(struct conhookmessage *, 	chm,		A1)
		 )
{
	AROS_USERFUNC_INIT

	struct MsgPort		*port, replyport;
	struct MyEditHookMsg	msg;

	D(bug("ConClip/conclipsnipfunc:\n"));
	D(bug("ConClip/conclipsnipfunc: snipdata 0x%lx sniplen %ld\n",
		  chm->chm_snipdata,
		  chm->chm_sniplen));

	if ((port = FindPort(CONCLIP_PORTNAME)))
	{
		replyport.mp_Node.ln_Type	=	NT_MSGPORT;
		replyport.mp_Node.ln_Name =	NULL;
		replyport.mp_Node.ln_Pri 	=	0;
		replyport.mp_Flags 	=	PA_SIGNAL;
		replyport.mp_SigBit 	=	SIGB_SINGLE;
		replyport.mp_SigTask 	=	FindTask(NULL);
		NewList(&replyport.mp_MsgList);

		msg.msg.mn_Node.ln_Type 	=	NT_MESSAGE;
		msg.msg.mn_ReplyPort 	=	&replyport;
		msg.msg.mn_Length 	=	sizeof(msg);

		msg.code			=	SNIPCODE_COPY;
		msg.sgw			=	NULL;
		msg.snipdata		=	chm->chm_snipdata;
		msg.sniplength		=	chm->chm_sniplen;

		D(bug("ConClip/conclipeditfunc: Msg Port 0x%lx\n", port));
		SetSignal(0, SIGF_SINGLE);
		PutMsg(port, &msg.msg);
		WaitPort(&replyport);
	} /* if ((port = FindPort(CONCLIP_PORTNAME))) */

	D(bug("ConClip/conclipsnipfunc: done\n"));
	AROS_USERFUNC_EXIT

	return (0);
}

/*****************************************************************************************/

static void installsniphook(void)
{
	D(bug("installsniphook:\n"));

	if (ConsoleDevice==NULL)
	{
		if (OpenDevice("console.device",
					   CONU_LIBRARY,
					   (struct IORequest*) &snipio,
					   0)==0)
		{
			ConsoleDevice =(struct Library*) snipio.io_Device;
		}
	}
	if (ConsoleDevice)
	{
		sniphook.h_Entry = (HOOKFUNC)AROS_ASMSYMNAME(conclipsnipfunc);
		sniphook.h_SubEntry = NULL;
		sniphook.h_Data = NULL;

		AddSnipHook(&sniphook);
		D(bug("installsniphook: done\n"));
	}
	else
	{
		D(bug("installsniphook: no ConsoleDevice\n"));
	}
}

/*****************************************************************************************/

static void removesniphook(void)
{
	if (ConsoleDevice)
	{
		D(bug("removesniphook:\n"));
		RemSnipHook(&sniphook);
		D(bug("removesniphook: done\n"));
		CloseDevice((struct IORequest*) &snipio);
		ConsoleDevice = NULL;
	}
	else
	{
		D(bug("removesniphook: no ConsoleDevice\n"));
	}
}




/*****************************************************************************************/

static void readfromclipboard(struct SGWork *sgw)
{
	struct IFFHandle    *iff;
	struct ContextNode  *cn;

	if((iff = AllocIFF()))
	{
		D(bug("ConClip/conclipeditfunc: AllocIFF okay\n"));

		if((iff->iff_Stream = (IPTR)OpenClipboard(clipunit)))
		{
			D(bug("ConClip/conclipeditfunc: OpenClipboard okay\n"));

			InitIFFasClip(iff);
			if(!OpenIFF(iff, IFFF_READ))
			{
				D(bug("ConClip/conclipeditfunc: OpenIff okay\n"));

				if (!(StopChunk(iff, ID_FTXT, ID_CHRS)))
				{
					D(bug("ConClip/conclipeditfunc: StopChunk okay\n"));

					if (!ParseIFF(iff, IFFPARSE_SCAN))
					{
						D(bug("ConClip/conclipeditfunc: ParseIFF okay\n"));

						cn = CurrentChunk(iff);
						if ((cn->cn_Type == ID_FTXT) && (cn->cn_ID == ID_CHRS) && (cn->cn_Size > 0))
						{
							WORD readsize;

							D(bug("ConClip: readfromclipboard: Found FTXT CHRS Chunk\n"));
							D(bug("ConClip: readfromclipboard: Old text = \"%s\"\n", sgw->WorkBuffer));

							readsize = sgw->StringInfo->MaxChars - 1 - sgw->BufferPos;
							if (cn->cn_Size < readsize) readsize = cn->cn_Size;
							if (readsize > 0)
							{
								memmove(sgw->WorkBuffer + sgw->BufferPos + readsize,
										sgw->WorkBuffer + sgw->BufferPos,
										sgw->StringInfo->MaxChars - sgw->BufferPos - readsize);
								ReadChunkBytes(iff, sgw->WorkBuffer + sgw->BufferPos, readsize);

								D(bug("ConClip: readfromclipboard: New text = \"%s\"\n", sgw->WorkBuffer));

								sgw->BufferPos += readsize;
							}

						} /* if ((cn->cn_Type == ID_FTXT) && (cn->cn_ID == ID_CHRS) && (cn->cn_Size > 0)) */

					} /* if (!ParseIFF(iff, IFFPARSE_SCAN)) */

				} /* if (!(StopChunk(iff, ID_FTXT, ID_CHRS))) */

				CloseIFF(iff);

			} /* if(!OpenIFF(iff, IFFF_READ)) */
			CloseClipboard((struct ClipboardHandle*)iff->iff_Stream);

		} /* if((iff->iff_Stream = (IPTR)OpenClipboard(clipunit))) */
		FreeIFF(iff);

	} /* if((iff = AllocIFF()))) */
}

/*****************************************************************************************/

static void ConClipHandler(void)
{
	BOOL quitme = FALSE;
	ULONG sigs;

	D(bug("ConClipHandler:\n"));

	if (DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37))
	{
		if (IFFParseBase=(struct Library *)OpenLibrary("iffparse.library",37))
		{
			if (UtilityBase=OpenLibrary("utility.library",37))
			{
				if (IntuitionBase=(struct IntuitionBase *)OpenLibrary("intuition.library",37))
				{
					if (createport())
					{
						installedithook();
						installsniphook();

						while(!quitme)
						{
							D(bug("ConClipHandler: portmask 0x%lx\n",
								  portmask));
							sigs = Wait(SIGBREAKF_CTRL_C | portmask);

							D(bug("ConClipHandler: receivedsigs 0x%lx\n",
								  sigs));

							if (sigs & portmask)
							{
								struct MyEditHookMsg *msg;

								while((msg = (struct MyEditHookMsg *)GetMsg(progport)))
								{
									switch(msg->code)
									{
									case EDITCODE_COPY:
										D(bug("ConClip: Received EDITCODE_COPY message\n"));
										savetoclipboard(msg->sgw->WorkBuffer,
														msg->sgw->NumChars);
										break;

									case EDITCODE_PASTE:
										D(bug("ConClip: Received EDITCODE_PASTE message\n"));
										readfromclipboard(msg->sgw);
										break;

									case SNIPCODE_COPY:
										D(bug("ConClip: Received SNIPCODE_COPY message\n"));
										savetoclipboard(msg->snipdata,
														msg->sniplength);
										break;

									default:
										D(bug("ConClip: Unknown Code 0x%lx\n",
											  msg->code));
										readfromclipboard(msg->sgw);
										break;
									} /* switch(msg->code) */
									ReplyMsg(&msg->msg);
								} /* while((msg = (struct MyEditHookMsg *)GetMsg(progport))) */
							} /* if (sigs & portmask) */

							if (sigs & SIGBREAKF_CTRL_C)
							{
								quitme = TRUE;
							}
						} /* while(!quitme) */
						if (oldedithook)
						{
							SetEditHook(oldedithook);
						}
						removesniphook();
						DeletePort(progport);
					}
					else
					{
						D(bug("ConClipHandler: can't create %s port\n",
							  CONCLIP_PORTNAME));
					}
					CloseLibrary((struct Library *)IntuitionBase);
				}
				else
				{
					D(bug("ConClipHandler: can't open intuition.library\n"));
				}
				CloseLibrary(UtilityBase);
			}
			else
			{
				D(bug("ConClipHandler: can't open utility.library\n"));
			}
			CloseLibrary((struct Library *)IFFParseBase);
		}
		else
		{
			D(bug("ConClipHandler: can't open iffparse.library\n"));
		}
		CloseLibrary((struct Library *)DOSBase);
	}
	else
	{
		D(bug("ConClipHandler: can't open dos.library\n"));
	}
	Forbid();
	UnLoadSeg(OldSegList);
}


/*****************************************************************************************/
